class Level {

	int[][] cells; 

	PImage level;
	PImage imageData;

	int gridWidth;
	int gridHeight;

	PVector scroll;

	int probabilityOfAliveAtStart = 25;

	float timeSpawnStarted = 0;
	float timeSpawnDelay = 10000;

	ArrayList<Cell> lastLazerCells;

	int countMaxBadCells = 20;
	int countMinBadCells = 4;

	boolean spawned = true;

	int timeSpawnDelayAFK = 20000;

	Level() {


		imageData = loadImage("data/corps.png");

		BuildGrid();

		DrawLevel();

		scroll = new PVector(screenX * -0.2, screenY * -0.55);

		lastLazerCells = new ArrayList<Cell>();
	}

	void BuildGrid() {

		imageData.loadPixels();

		gridWidth = int(imageData.width/cellSize);
		gridHeight = int(imageData.height/cellSize);

  		cells = new int[gridWidth][gridHeight];

  		level = createImage(imageData.width, imageData.height, RGB);
  		level.loadPixels();

	    int pixelCount = imageData.pixels.length;

	    ResetCount();

		for (int p = 0; p < pixelCount; p++) {
			int x = (p % imageData.width) / cellSize;
			int y = floor((p / imageData.width) / cellSize);
			int cellStatus = GetTypeByColor(imageData.pixels[p]);
			cells[x][y] = cellStatus;
			CountCell(cellStatus);
		}

		CheckoutCount();
	}

	void Update() {
		image(level, scroll.x, scroll.y);
		// DrawPlayer();
		DrawBullets();

		// Spawn new wave of Bad Cells
		if (timeSpawnStarted + timeSpawnDelay < millis()) {
		// || timeSpawnStarted + timeSpawnDelayAFK < millis()) {
			SpawnBad();
			DrawLevel();
			// timeSpawnDelay = 10000 + random(3000);
			// spawned = true;
			timeSpawnStarted = millis();
			waves++;
		}

		// if (spawned && cellCountBad < countMinBadCells) {
		// 	timeSpawnStarted = millis();
		// 	spawned = false;
		// }
	}

	String[] automataObject = {"glider.png", "puffer0.png", "puffer2.png", "spaceship_heavy.png", "spaceship_light.png", "spaceship_middle.png"};

	void SpawnObject(int x, int y) {
		String randomObject = automataObject[int(random(automataObject.length))];
		PImage imageObject = loadImage("data/" + randomObject);
		imageObject.loadPixels();

		int pixelCount = imageObject.pixels.length;
		for (int p = 0; p < pixelCount; p++) {
			int iX = x + p % imageObject.width;
			int iY = y + floor(p / imageObject.width);
			if (IsPositionValid(iX, iY)) {
				cells[iX][iY] = cellStatusBad;
			}
		}

		println("spawn " + randomObject);
	}

	void SpawnBad() {

		int xx = int((random(screenX) + abs(scroll.x)) / cellSize);
		int yy = int((random(screenY) + abs(scroll.y)) / cellSize);

		if(random(100) < 50) {

			// Spawn object

			SpawnObject(xx, yy);

		} else {

			// Random square

			int offset = 10 + int(random(10));
			int w = xx + offset;
			int h = yy + offset;
			for (int x=xx; x<w; x++) {
				for (int y=yy; y<h; y++) {
					if (IsPositionValid(x, y)) {
						float state = random (100);
						if (state <= probabilityOfAliveAtStart) {
							cells[x][y] = cellStatusBad;
						}
					}
				}
			}
		}
	}

	// void DrawPlayer() {
	//     int pixelCount = player.bodySize * player.bodySize;

	//     PVector playerPos = new PVector(player.position.x, player.position.y);

	//     fill(player.cellColor);

	// 	for (int p = 0; p < pixelCount; p++) {
	// 		int iX = (p % player.bodySize);
	// 		int iY = floor(p/player.bodySize);
	// 		int x = int(playerPos.x + iX * cellSize);
	// 		int y = int(playerPos.y + iY * cellSize);
	// 		color cellColor = player.body[iX][iY].cellColor;
	// 		rect(x, y, cellSize, cellSize);
	// 	}
	// }

	void DrawBullets() {

		for (int b = 0; b < player.bullets.size(); b++) {

			Bullet bullet = player.bullets.get(b);

			if (bullet.IsAlive()) {

				bullet.position.x += bullet.direction.x * bullet.speed;
				bullet.position.y += bullet.direction.y * bullet.speed;

				if (IsPositionValid(bullet.x, bullet.y))
					cells[bullet.x][bullet.y] = bullet.previousCellStatus;

				int x = int(bullet.position.x + abs(scroll.x)) / cellSize;
				int y = int(bullet.position.y + abs(scroll.y)) / cellSize;

				if (IsPositionValid(x, y)) {

					bullet.x = x;
					bullet.y = y;
					bullet.previousCellStatus = cells[x][y];

					// Bullet out of Screen
					if (x < 0 || x >= gridWidth || y < 0 || y >= gridHeight ||
						// Out of time
						bullet.timeStarted + bullet.timeDelay < millis()) {

						bullet.Kill();

					// Bullet on screen				
					} else {

						int currentCell = cells[x][y];

						// Check for bad neighbours
						ArrayList<Cell> neighbours = GetNeighbours(x, y, cellStatusBad);

						// Kill bad neighbours
						int badCount = neighbours.size();
						if (badCount > 0) {
							bullet.Kill();
							for (int n = 0; n < badCount; n++) {
								Cell cellNeighbour = neighbours.get(n);
								int nX = int(cellNeighbour.x);
								int nY = int(cellNeighbour.y);
								cells[nX][nY] = cellStatusDead;
							}
						}
						// Bullet cell /* COMMENT FOR DIRTY MODE */
						else if (currentCell == cellStatusBullet) {
							bullet.Kill();
						}
						// 
						else if (currentCell == cellStatusDead) {

							/* HEAL */

							int neighboursVital = HowManyVitalNeighbours(x, y);
							if (neighboursVital > 0) {
								int averageNeighbour = GetAverageNeighbour(x, y);
								// println("averageNeighbour: "+averageNeighbour);
								bullet.Kill();
								SetNeighbours(x, y, averageNeighbour);
								cells[x][y] = averageNeighbour;
							} else {
								// bullet.AddGhost(x, y, currentCell);
								cells[x][y] = cellStatusBullet;
							}

						} else {

							// bullet.AddGhost(x, y, currentCell);
							cells[x][y] = cellStatusBullet;
						}

					}
				}
			}
		}
		
	}

	void DrawLevel() {

	    level.loadPixels();

	    int pixelCount = level.pixels.length;

	    ResetCount();

		for (int p = 0; p < pixelCount; p++) {
			int x = (p % imageData.width) / cellSize;
			int y = floor((p/imageData.width) / cellSize);
			level.pixels[p] = GetColorByType(cells[x][y]);
			CountCell(cells[x][y]);
		}
	    level.updatePixels();
	}

	void CleanLazer() {
		for (int c = 0; c < lastLazerCells.size(); c++) {
			Cell cell = lastLazerCells.get(c);
			cells[cell.x][cell.y] = cell.status;
		}
		lastLazerCells = new ArrayList<Cell>();
	}

	void FireLazer() {
		CleanLazer();
		int fromX = int((player.position.x - scroll.x) / cellSize);
		int fromY = int((player.position.y - scroll.y) / cellSize);
		int toX = int((mouseX - scroll.x) / cellSize);
		int toY = int((mouseY - scroll.y) / cellSize);
		float t = 0.0;
		while(t < 1.0) {
			int x = int(fromX + (toX - fromX) * t);
			int y = int(fromY +  (toY - fromY) * t);
			if (IsPositionValid(x, y) && cells[x][y] != cellStatusBullet) {
				lastLazerCells.add(new Cell(x, y, cells[x][y]));
				cells[x][y] = cellStatusBullet;
			}
			t += 0.05;
		}
	}

	boolean IsPositionValid(int xx, int yy) {
		return ((xx>=0)&&(xx<imageData.width/cellSize))&&((yy>=0)&&(yy<imageData.height/cellSize));
	}

	boolean IsPositionValid(int xx, int yy, int ww, int hh) {
		return ((xx>=0)&&(xx<ww/cellSize))&&((yy>=0)&&(yy<hh/cellSize));
	}

	int HowManyNeighbours(int x, int y, int cellStatus) { 
		int count = 0;
		for (int xx=x-1; xx<=x+1;xx++)
			for (int yy=y-1; yy<=y+1;yy++)
				if (IsPositionValid(xx, yy)) 
					if (!((xx==x)&&(yy==y)))
						if (cells[xx][yy] == cellStatus)
							count++;
		return count;
	}

	ArrayList<Cell> GetNeighbours(int x, int y, int cellStatus) { 
		ArrayList<Cell> neighbours = new ArrayList<Cell>();
		for (int xx=x-1; xx<=x+1;xx++)
			for (int yy=y-1; yy<=y+1;yy++)
				if (IsPositionValid(xx, yy)) 
					if (!((xx==x)&&(yy==y)))
						if (cells[xx][yy] == cellStatus)
							neighbours.add(new Cell(xx, yy, cellStatus));
		return neighbours;
	}

	int HowManyVitalNeighbours(int x, int y) {
		int vitals = 0;
		for (int xx=x-1; xx<=x+1;xx++) {
			for (int yy=y-1; yy<=y+1;yy++) {
				if (IsPositionValid(xx, yy)) {
					if (!((xx==x)&&(yy==y))) {
						int cell = cells[xx][yy];
						if (cell != cellStatusDead && cell != cellStatusBullet && cell != cellStatusBad) {
							vitals++;
						}
					}
				}
			}
		}
		return vitals;
	}

	int GetAverageNeighbour(int x, int y) { 
		int firstColor = 0;
		IntList neighbours = new IntList();
		for (int xx=x-1; xx<=x+1;xx++) {
			for (int yy=y-1; yy<=y+1;yy++) {
				if (IsPositionValid(xx, yy)) {
					if (!((xx==x)&&(yy==y))) {
						int cell = cells[xx][yy];
						if (cell != cellStatusDead && cell != cellStatusBullet) {
							if (neighbours.hasValue(cell)) {
								return cell;
							} else {
								neighbours.append(cell);
								if (firstColor == 0) firstColor = cell;
							}
						}
					}
				}
			}
		}
		return firstColor;
	}

	void SetNeighbours(int x, int y, int cellStatus) {
		for (int xx=x-1; xx<=x+1;xx++)
			if (IsPositionValid(xx, y)) 
				if (!(xx==x))
					cells[xx][y] = cellStatus;
		for (int yy=y-1; yy<=y+1;yy++)
			if (IsPositionValid(x, yy)) 
				if (!(yy==y))
					cells[x][yy] = cellStatus;
	}
}
