int screenX = 800;
int screenY = 800;

int scrollOffset = 128;

// How likely for a cell to be alive at start (in percentage)
float probabilityOfAliveAtStart = 15;

// Variables for timer
int interval = 50;
int lastRecordedTime = 0;

int waves = 0;

// Menu vars
// boolean playing = false;
boolean pause = true;
boolean gameWin = false;
boolean gameOver = false;

Input input;
Player player;
Level level;
AutomataCellular automataCellular;
Hud hud;

int cellSize = 4;
int newCellSize = 4;

void setup() {
  size (screenX, screenY);
  noSmooth();

  Init();
  Clean();
}

void Clean() {
  noStroke();
  background(0);
}

void draw() {

  Clean();

  if (!pause) {
    UpdateGame();
  }

  UpdateMenu();

  if (newCellSize != cellSize) {
    cellSize = newCellSize;
    StartGame();
    pause = false;
  }
}

void UpdateMenu() {
  hud.Update();
}

void StartGame() {
  Init();
}

void Init() {
  pause = true;
  gameOver = false;
  gameWin = false;
  ResetCount();
  waves = 0;

  input = new Input();
  player = new Player();
  level = new Level(); 
  automataCellular = new AutomataCellular();
  hud = new Hud();
}

void UpdateGame() {

  // Iterate if timer ticks
  if (millis()-lastRecordedTime>interval) {
    automataCellular.iteration(level.cells);
    lastRecordedTime = millis();
    level.DrawLevel();
  }

  level.Update();

  player.Update();

  CheckGameOver();

  if (mousePressed && (mouseButton == RIGHT)) {
    level.FireLazer();
  }

}

void CheckGameOver() {
  // println("ratio : "+cellCountVital / float(cellTotalVital) );
  //  ||
    //GetBadCellRatio() > 0.5
  if (GetVitalCellRatio() < 0.1) {
    gameOver = true;
    pause = true;
  }
}

void keyReleased() {

  input.KeyReleased(key, keyCode);

  // DEBUG
  if (key == 'o') level.SpawnBad();
  // if (key == 'o') gameOver = true;
  if (key == 'r') StartGame();
  if (key == '+') newCellSize = 8;
  if (key == '-') newCellSize = 4;
}

void keyPressed() {

  input.KeyPressed(key, keyCode);

}

void mouseReleased() {
  
  level.CleanLazer();

}
